@@ -10,7 +10,10 @@ import matplotlib.ticker as ticker
 import numpy as np
 import pandas as pd
 import seaborn as sns
+from filterpy.kalman import KalmanFilter
+from scipy.stats import kurtosis
 from scipy.stats import pearsonr
+from scipy.stats import skew as sskew
 from sklearn import metrics
 from sklearn import preprocessing
 from sklearn.cluster import KMeans
@@ -42,20 +45,28 @@ class plot:
     """
     data必须是dataframe,column_name是特征名字
     """
-    def histplot(data, column_name, **kwargs):
+    def histplot(data, column_name, ifskew=False, **kwargs):
         try:
             skew = data[column_name].skew()
             kurt = data[column_name].kurt()
             X = np.array(data[column_name])
+            bins = 0.1
+            if ifskew:
+                X = np.log1p(X)
+                bins = 0.005
+                skew = sskew(X)
+                kurt = kurtosis(X)
             X_min = min(X)
             X_max = max(X)
             X_range = X_max - X_min
             X_min = int(kwargs['xmin']) if 'xmin' in kwargs else min(X) - X_range *0.05
             X_max = int(kwargs['xmax']) if 'xmax' in kwargs else max(X) + X_range *0.05
-            column_bins = np.arange(np.min(X), np.max(X), 0.1)
+            column_bins = np.arange(np.min(X), np.max(X), bins)
             column_mean = np.mean(X)
             column_std = np.std(X)
             plt.figure(figsize=(16, 8), dpi=100)
+            plt.xticks(fontsize=14)
+            plt.yticks(fontsize=14)
             plt.xlim(X_min, X_max)
             sns.distplot(X, kde=True, norm_hist=True, bins=column_bins, color="black")
             plt.title(column_name + " mean=%.1f std=%.1f skew=%.2f kurt=%.2f"
@@ -88,6 +99,7 @@ class plot:
                 raise Exception(err)
 
             perplexity = int(kwargs['perplexity']) if 'perplexity' in kwargs else 30
+            Bins = float(kwargs['bin_width']) if 'bin_width' in kwargs else 0.1
             X = np.array(data[column_names])
             fname = []
             score_silhouette_list = []
@@ -99,7 +111,8 @@ class plot:
                     X_range = X_max - X_min
                     X_min = int(kwargs['xmin']) if 'xmin' in kwargs else min(X) - X_range * 0.05
                     X_max = int(kwargs['xmax']) if 'xmax' in kwargs else max(X) + X_range * 0.05
-                    feature_bins = np.arange(np.min(X), np.max(X), 0.1)
+                    Y_max = int(kwargs['ymax']) if 'ymax' in kwargs else 0
+                    feature_bins = np.arange(np.min(X), np.max(X), Bins)
 
                     X_array = X.reshape(len(X), 1)
                     gmm = GaussianMixture(n_components=N_COMPONENTS).fit(X_array)
@@ -116,7 +129,11 @@ class plot:
                     score_silhouette_list.append(score_silhouette)
                     score_calinski_list.append(score_calinski)
                     plt.figure(figsize=(16, 8), dpi=100)
+                    plt.xticks(fontsize=14)
+                    plt.yticks(fontsize=14)
                     plt.xlim(X_min, X_max)
+                    if Y_max != 0:
+                        plt.ylim(0, Y_max)
                     plt.title("GMM: K=%d score_silhouette=%.3f Tsen_perplexity=%d" %
                               (N_COMPONENTS, score_silhouette, perplexity))
                     for k in range(N_COMPONENTS):
@@ -145,6 +162,8 @@ class plot:
                     score_silhouette_list.append(score_silhouette)
                     score_calinski_list.append(score_calinski)
                     plt.figure(figsize=(16, 8), dpi=100)
+                    plt.xticks(fontsize=14)
+                    plt.yticks(fontsize=14)
                     plt.title("GMM: K=%d score_silhouette=%.3f" % (N_COMPONENTS, score_silhouette))
                     X_tsen = X
                     if len(column_names) > 2:
@@ -171,16 +190,20 @@ class plot:
                 print("K=%d\n%s" % (row[1], row[0]))
 
             plt.figure(figsize=(16, 8), dpi=100)
+            ax = plt.subplot()
+            plt.xticks(fontsize=14)
+            plt.yticks(fontsize=14)
             plt.title("GMM silhouette & calinski_harabasz")
             score_silhouette = np.array(score_silhouette_list)
             score_silhouette_max = np.max(score_silhouette)
             score_calinski = np.array(score_calinski_list)
             score_calinski_max = np.max(score_calinski)
-            p1, = plt.plot(range(k_range[0], k_range[1]), score_silhouette / score_silhouette_max, marker='o')
-            p2, = plt.plot(range(k_range[0], k_range[1]), score_calinski / score_calinski_max, marker='o')
+            p1, = ax.plot(range(k_range[0], k_range[1]), score_silhouette / score_silhouette_max, marker='o')
+            p2, = ax.plot(range(k_range[0], k_range[1]), score_calinski / score_calinski_max, marker='o')
             plt.legend([p1, p2], ["score_silhouette MAX=%.2f" % score_silhouette_max,
                                   "score_calinski MAX=%.2f" % score_calinski_max], loc='upper left')
             datetime_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S%f')
+            ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
             file_name = hashlib.md5(datetime_str.encode(encoding='UTF-8')).hexdigest()
             file_name = file_name + ".png"
             print("GMM score")
@@ -196,6 +219,7 @@ class plot:
     def kmeansplot(data, column_names, k_range, **kwargs):
         try:
             perplexity = int(kwargs['perplexity']) if 'perplexity' in kwargs else 30
+            Bins = float(kwargs['bin_width']) if 'bin_width' in kwargs else 0.1
             X = np.array(data[column_names])
             iteration_info = []
             score_silhouette_list = []
@@ -208,7 +232,8 @@ class plot:
                     X_range = X_max - X_min
                     X_min = int(kwargs['xmin']) if 'xmin' in kwargs else min(X) - X_range * 0.05
                     X_max = int(kwargs['xmax']) if 'xmax' in kwargs else max(X) + X_range * 0.05
-                    feature_bins = np.arange(np.min(X), np.max(X), 0.1)
+                    Y_max = int(kwargs['ymax']) if 'ymax' in kwargs else 0
+                    feature_bins = np.arange(np.min(X), np.max(X), Bins)
 
                     X_array = X.reshape(len(X), 1)
 
@@ -235,6 +260,10 @@ class plot:
                     inertia_list.append(s.inertia_)
                     plt.figure(figsize=(16, 8), dpi=100)
                     plt.xlim(X_min, X_max)
+                    if Y_max != 0:
+                        plt.ylim(0, Y_max)
+                    plt.xticks(fontsize=14)
+                    plt.yticks(fontsize=14)
                     plt.title("Kmeans: K=%d score_silhouette=%.3f inertia=%.3f" %
                               (N_CLUSTERS, score_silhouette, s.inertia_))
                     for k in range(N_CLUSTERS):
@@ -267,6 +296,8 @@ class plot:
                     score_calinski_list.append(score_calinski)
                     inertia_list.append(s.inertia_)
                     plt.figure(figsize=(16, 8), dpi=100)
+                    plt.xticks(fontsize=14)
+                    plt.yticks(fontsize=14)
                     plt.title("Kmeans: K=%d score_silhouette=%.3f inertia=%.3f" %
                               (N_CLUSTERS, score_silhouette, s.inertia_))
                     X_tsen = X
@@ -295,6 +326,9 @@ class plot:
                 print(row[2])
 
             plt.figure(figsize=(16, 8), dpi=100)
+            ax = plt.subplot()
+            plt.xticks(fontsize=14)
+            plt.yticks(fontsize=14)
             plt.title("GMM silhouette & calinski_harabasz & inertia")
             score_silhouette = np.array(score_silhouette_list)
             score_silhouette_max = np.max(score_silhouette)
@@ -302,13 +336,14 @@ class plot:
             score_calinski_max = np.max(score_calinski)
             inertia = np.array(inertia_list)
             inertia_max = np.max(inertia)
-            p1, = plt.plot(range(k_range[0], k_range[1]), score_silhouette / score_silhouette_max, marker='o')
-            p2, = plt.plot(range(k_range[0], k_range[1]), score_calinski / score_calinski_max, marker='o')
-            p3, = plt.plot(range(k_range[0], k_range[1]), inertia / inertia_max, marker='o')
+            p1, = ax.plot(range(k_range[0], k_range[1]), score_silhouette / score_silhouette_max, marker='o')
+            p2, = ax.plot(range(k_range[0], k_range[1]), score_calinski / score_calinski_max, marker='o')
+            p3, = ax.plot(range(k_range[0], k_range[1]), inertia / inertia_max, marker='o')
             plt.legend([p1, p2, p3], ["score_silhouette MAX=%.2f" % score_silhouette_max,
                                       "score_calinski MAX=%.2f" % score_calinski_max,
                                       "inertia MAX=%.2f" % inertia_max], loc='upper left')
             datetime_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S%f')
+            ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
             file_name = hashlib.md5(datetime_str.encode(encoding='UTF-8')).hexdigest()
             file_name = file_name + ".png"
             print("Kmeans score")
@@ -529,6 +564,81 @@ class plot:
             print(err)
             return False
 
+    def kalmanplot(data, column_names, dim_x=2, dim_z=1, x=[], p=[], f=[], q=[], h=[], r=1, **kwargs):
+        try:
+            data = np.array(data[column_names])
+
+            length = len(column_names)
+            if length == 1:
+                if not x:
+                    x = [1, 0.1]
+                if not p:
+                    p = [[1, 0.1], [0.1, 1]]
+                if not f:
+                    f = [[1, 0.5], [0, 1]]
+                if not q:
+                    q = [[0.0001, 0], [0, 0.0001]]
+                if not h:
+                    h = [[1, 0]]
+
+            elif length > 1:
+                if not x or not p or not f or not q or not h:
+                    err = "when length of column is larger than 1, the (x, p, f, q, h) matrixs cannot be default"
+                    raise Exception(err)
+                if length != len(x) or length != len(p) or length != len(f) or length != len(q):
+                    err = "matrixs should have same dimension with colums_names"
+                    raise Exception(err)
+
+            # dim_x为状态空间维度, z为观测矩阵维度
+            kf = KalmanFilter(dim_x=dim_x, dim_z=dim_z)
+            # 心率变化率为0.1是因为每5s心率变化不会太大
+            kf.x = np.array(x)
+            # 心率变化率和人的心率是一定的关系，根据运动状态或者濒死会有很明显的差别
+            # 0.1代表有一定关系，但是不大
+            kf.P = np.array(p)
+            # F不能设置为两个0.5，此矩阵不能对称，否则结果异常
+            kf.F = np.array(f)
+            # 噪音，目前数据都是cpu运算的，并不会有噪音
+            # Q不能设置为0，为0结果会异常
+            kf.Q = np.array(q)
+            # 观测矩阵，我们只观测心率，并不能观察变化率
+            # Sometimes certain states are measured, when others are not.
+            # For example, the first, third and fifth states of a five-dimensional state vector are measurable,
+            # while second and fourth states are not measurable
+            # H = [1 0 0 0 0][0 0 1 0 0][0 0 0 0 1]
+            kf.H = np.array(h)
+            # 测量误差，医疗器械心率误差规定为+-1
+            kf.R = r
+
+            newList = []
+            for i in range(len(data)):
+                kf.predict()
+                kf.update(data[i])
+                x = kf.x
+                newList.append(x)
+            newList = np.array(newList)
+
+            for i in range(len(data[0])):
+                plt.figure(figsize=(32, 8), dpi=100)
+                p1 = plt.scatter(x=range(len(data)), y=data[:, i], marker=".", color="r")
+                p2 = plt.scatter(x=range(len(newList)), y=newList[:, i], marker=".", color="b")
+                # p1, = plt.plot(x, color="r")
+                # p2, = plt.plot(x1, color="b")
+                plt.legend([p1, p2], ["Original Data", "KalmanFilter Data"], loc='upper left')
+
+                datetime_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S%f')
+                file_name = hashlib.md5(datetime_str.encode(encoding='UTF-8')).hexdigest()
+                file_name = file_name + ".png"
+                print(file_name)
+                plt.savefig(file_name, dpi=100)
+                return True
+
+        except Exception as err:
+            print("An error occured!")
+            print(err)
+            return False
+
+
 
 class seaborn:
     def __init__(self):
